use strict;
use warnings;
use Test::More;

# generated by Dist::Zilla::Plugin::Test::PodSpelling 2.007005
use Test::Spelling 0.12;
use Pod::Wordlist;


add_stopwords(<DATA>);
all_pod_files_spelling_ok( qw( bin lib ) );
__DATA__
Alders
Alders'
Bieron
CIDR
DB
Dave
Eilam
Eilam's
Florian
Fowler
GeoIP
Greg
Ilya
Inc
Jan
MAXMIND
MAXMIND's
MMDB
Mark
Mather
MaxMind
MaxMind's
Melamed
Olaf
Oschwald
Oschwald's
PayPal
Processor
Ragwitz
Rolsky
Rolsky's
Serializer
Storey
TW
Teredo
Thomas
Tree
Util
VisualizeTree
Will
William
Writer
docs
drolsky
en
goschwald
ilya77
jbieron
lib
mfowler
oalders
rafl
routable
tjmather
toplevel
uint
will
wstorey
zh
